"""
Author: Mingjian He <mh1@stanford.edu>

switching module contains switching state-space models in SOMATA
"""

from .traditional import switching  # noqa: F401
from .vb import VBSwitchModel  # noqa: F401
